/*
 *   Copyright (C) 2009,2010 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "OpenG2ProtocolHandler.h"

#include "CCITTChecksum.h"
#include "DStarDefines.h"
#include "Utils.h"

// #define	DUMP_TX

const unsigned char BAND_2M   = 0x02U;
const unsigned char BAND_70CM = 0x01U;
const unsigned char BAND_23CM = 0x03U;
const unsigned char BAND_10M  = 0x04U;
const unsigned char BAND_DD   = 0x00U;

const unsigned int BUFFER_LENGTH = 255U;

COpenG2ProtocolHandler::COpenG2ProtocolHandler(const wxString& gatewayAddress, unsigned int gatewayPort, const wxString& localAddress, unsigned int localPort) :
m_socket(gatewayAddress, gatewayPort, localAddress, localPort),
m_radioCallsign(NULL),
m_controllerCallsign(NULL),
m_band(BAND_2M),
m_outId(0U),
m_outSeq1(0U),
m_outSeq2(0U),
m_type(NETWORK_NONE),
m_inId(0U),
m_inSeq1(0U),
m_userSeq2(0U),
m_statusSeq2(0U),
m_buffer(NULL),
m_length(0U)
{
	m_radioCallsign      = new unsigned char[LONG_CALLSIGN_LENGTH];
	m_controllerCallsign = new unsigned char[LONG_CALLSIGN_LENGTH];

	::memset(m_radioCallsign,      ' ', LONG_CALLSIGN_LENGTH);
	::memset(m_controllerCallsign, ' ', LONG_CALLSIGN_LENGTH);

	m_buffer = new unsigned char[BUFFER_LENGTH];

	wxDateTime now = wxDateTime::Now();
	::srand(now.GetMillisecond());
}

COpenG2ProtocolHandler::~COpenG2ProtocolHandler()
{
	delete[] m_radioCallsign;
	delete[] m_controllerCallsign;
	delete[] m_buffer;
}

void COpenG2ProtocolHandler::setRepeaterCallsign(const wxString& callsign)
{
	for (unsigned int i = 0U; i < callsign.Len(); i++)
		m_radioCallsign[i] = callsign.GetChar(i);

	for (unsigned int i = 0U; i < callsign.Len(); i++)
		m_controllerCallsign[i] = callsign.GetChar(i);

	// Replace the last letter with an S
	m_controllerCallsign[7U] = 'S';

	switch (m_radioCallsign[7U]) {
		case 'A':
			m_band = BAND_23CM;
			break;
		case 'B':
			m_band = BAND_70CM;
			break;
		case 'C':
			m_band = BAND_2M;
			break;
		case 'D':
			m_band = BAND_10M;
			break;
		default:
			m_band = BAND_70CM;
			break;
	}
}

bool COpenG2ProtocolHandler::open()
{
	return m_socket.open();
}

bool COpenG2ProtocolHandler::writeHeader(const CHeaderData& header)
{
	unsigned char buffer[58U];

	buffer[0] = 'D';
	buffer[1] = 'S';
	buffer[2] = 'T';
	buffer[3] = 'R';

	m_outSeq1++;
	buffer[4] = m_outSeq1 / 256U;	// Packet sequence number
	buffer[5] = m_outSeq1 % 256U;

	buffer[6] = 0x73;				// Not a response
	buffer[7] = 0x12;				// Data type

	buffer[8] = 0x00;				// Length of 48 bytes following
	buffer[9] = 0x30;

	buffer[10] = 0x20;				// AMBE plus Slow Data following

	buffer[11] = 0x00;				// Repeater to OpenG2 Gateway
	buffer[12] = 0x01;
	buffer[13] = m_band;

	// Create a random id for this transmission
	m_outId = (::rand() % 65535U) + 1U;

	buffer[14] = m_outId / 256U;	// Unique session id
	buffer[15] = m_outId % 256U;

	buffer[16] = 0x80;

	buffer[17] = 0x00;				// Flags 1, 2 and 3
	buffer[18] = 0x00;
	buffer[19] = 0x00;

	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++)
		buffer[20 + i] = header.getRptCall1().GetChar(i);

	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++)
		buffer[28 + i] = header.getRptCall2().GetChar(i);

	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++)
		buffer[36 + i] = header.getYourCall().GetChar(i);

	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++)
		buffer[44 + i] = header.getMyCall1().GetChar(i);

	for (unsigned int i = 0U; i < SHORT_CALLSIGN_LENGTH; i++)
		buffer[52 + i] = header.getMyCall2().GetChar(i);

	// Get the checksum for the header
	CCCITTChecksum csum;
	csum.update(buffer + 17, 4U * LONG_CALLSIGN_LENGTH + SHORT_CALLSIGN_LENGTH + 3U);
	csum.result(buffer + 56);

	m_outSeq2 = 0U;

#if defined(DUMP_TX)
	CUtils::dump(wxT("Sending Header"), buffer, 58U);
	return true;
#else
	return m_socket.write(buffer, 58U);
#endif
}

bool COpenG2ProtocolHandler::writeData(const unsigned char* data, unsigned int length, bool end)
{
	wxASSERT(data != NULL);
	wxASSERT(length == DV_FRAME_LENGTH_BYTES);

	unsigned char buffer[32U];

	buffer[0] = 'D';
	buffer[1] = 'S';
	buffer[2] = 'T';
	buffer[3] = 'R';

	m_outSeq1++;
	buffer[4] = m_outSeq1 / 256U;	// Packet sequence number
	buffer[5] = m_outSeq1 % 256U;

	buffer[6] = 0x73;				// Request
	buffer[7] = 0x12;				// Data

	buffer[8] = 0x00;				// Length of 19 bytes following
	buffer[9] = 0x13;

	buffer[10] = 0x20;				// AMBE plus Slow Data following

	buffer[11] = 0x00;				// Repeater to OpenG2 Gateway
	buffer[12] = 0x01;
	buffer[13] = m_band;

	buffer[14] = m_outId / 256U;	// Unique session id
	buffer[15] = m_outId % 256U;

	// If this is a data sync, reset the sequence to zero
	if (data[9] == 0x55 && data[10] == 0x2D && data[11] == 0x16)
		m_outSeq2 = 0U;

	buffer[16] = m_outSeq2;

	// Copy the AMBE data
	buffer[17] = data[0];
	buffer[18] = data[1];
	buffer[19] = data[2];
	buffer[20] = data[3];
	buffer[21] = data[4];
	buffer[22] = data[5];
	buffer[23] = data[6];
	buffer[24] = data[7];
	buffer[25] = data[8];

	// Copy Slow Data or the first part of the end marker
	if (end) {
		buffer[26] = 0x55;
		buffer[27] = 0x55;
		buffer[28] = 0x55;
	} else if (m_outSeq2 == 0U) {
		buffer[26] = 0x55;
		buffer[27] = 0x2D;
		buffer[28] = 0x16;
	} else {
		buffer[26] = data[9];
		buffer[27] = data[10];
		buffer[28] = data[11];
	}

	m_outSeq2++;
	if (m_outSeq2 > 0x14U)
		m_outSeq2 = 0U;

#if defined(DUMP_TX)
	CUtils::dump(wxT("Sending Data"), buffer, 29U);
#else
	bool res = m_socket.write(buffer, 29U);
	if (!res)
		return false;
#endif

	if (!end)
		return true;

	buffer[0] = 'D';
	buffer[1] = 'S';
	buffer[2] = 'T';
	buffer[3] = 'R';

	m_outSeq1++;
	buffer[4] = m_outSeq1 / 256U;	// Packet sequence number
	buffer[5] = m_outSeq1 % 256U;

	buffer[6] = 0x73;				// Request
	buffer[7] = 0x12;				// Data

	buffer[8] = 0x00;				// Length of 22 bytes following
	buffer[9] = 0x16;

	buffer[10] = 0x20;				// AMBE plus Slow Data following

	buffer[11] = 0x00;				// Repeater to OpenG2 Gateway
	buffer[12] = 0x01;
	buffer[13] = m_band;

	buffer[14] = m_outId / 256U;	// Unique session id
	buffer[15] = m_outId % 256U;

	buffer[16] = m_outSeq2 | 0x40;	// End of data marker
	m_outSeq2++;

	// Copy the dummy AMBE and slow data
	buffer[17] = 0x00;
	buffer[18] = 0x00;
	buffer[19] = 0x00;
	buffer[20] = 0x00;
	buffer[21] = 0x00;
	buffer[22] = 0x00;
	buffer[23] = 0x00;
	buffer[24] = 0x00;
	buffer[25] = 0x00;
	buffer[26] = 0x00;
	buffer[27] = 0x00;
	buffer[28] = 0x00;

	// Copy the rest of the end marker
	buffer[29] = 0x25;
	buffer[30] = 0x1A;
	buffer[31] = 0xC6;

#if defined(DUMP_TX)
	CUtils::dump(wxT("Sending Data"), buffer, 32U);
	return true;
#else
	return m_socket.write(buffer, 32U);
#endif
}

bool COpenG2ProtocolHandler::writeHeard(const wxString& callsign)
{
	unsigned char buffer[26U];

	buffer[0] = 'D';
	buffer[1] = 'S';
	buffer[2] = 'T';
	buffer[3] = 'R';

	m_outSeq1++;
	buffer[4] = m_outSeq1 / 256U;	// Packet sequence number
	buffer[5] = m_outSeq1 % 256U;

	buffer[6] = 0x73;				// Request
	buffer[7] = 0x21;				// Heard data

	buffer[8] = 0x00;				// Length of 16 bytes following
	buffer[9] = 0x10;

	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++)
		buffer[10U + i] = callsign.GetChar(i);

	::memcpy(buffer + 18U, m_radioCallsign, LONG_CALLSIGN_LENGTH);

#if defined(DUMP_TX)
	CUtils::dump(wxT("Sending Heard"), buffer, 26U);
	return true;
#else
	return m_socket.write(buffer, 26U);
#endif
}

bool COpenG2ProtocolHandler::writePoll()
{
	unsigned char buffer[26U];

	buffer[0] = 'D';
	buffer[1] = 'S';
	buffer[2] = 'T';
	buffer[3] = 'R';

	m_outSeq1++;
	buffer[4] = m_outSeq1 / 256U;	// Packet sequence number
	buffer[5] = m_outSeq1 % 256U;

	buffer[6] = 0x73;				// Request
	buffer[7] = 0x21;				// Poll data

	buffer[8] = 0x00;				// Length of 16 bytes following
	buffer[9] = 0x10;

	::memcpy(buffer + 10U, m_controllerCallsign, LONG_CALLSIGN_LENGTH);

	::memcpy(buffer + 18U, m_controllerCallsign, LONG_CALLSIGN_LENGTH);

#if defined(DUMP_TX)
	CUtils::dump(wxT("Sending Poll"), buffer, 26U);
	return true;
#else
	return m_socket.write(buffer, 26U);
#endif
}

NETWORK_TYPE COpenG2ProtocolHandler::read()
{
	bool res = true;

	// Loop until we have no more data from the socket or we have data for the higher layers
	while (res)
		res = readPackets();

	return m_type;
}

bool COpenG2ProtocolHandler::readPackets()
{
	m_type = NETWORK_NONE;

	// No more data?
	m_length = m_socket.read(m_buffer, BUFFER_LENGTH);
	if (m_length == 0U)
		return false;

	// Invalid packet type?
	if (m_buffer[0] != 'D' || m_buffer[1] != 'S' || m_buffer[2] != 'T' || m_buffer[3] != 'R')
		return true;

	// Status data
	else if (m_buffer[6] == 0x73 && m_buffer[7] == 0x12 && m_buffer[10] == 0x20 && m_buffer[11] == 0xFF && m_buffer[12] == 0xFF && m_buffer[13] == 0xFF) {
		m_inSeq1     = m_buffer[4] * 256U + m_buffer[5];
		m_statusSeq2 = m_buffer[16] & 0x0F;

		if (m_buffer[16] == 0x80)
			m_type = NETWORK_STATUS_HEADER;
		else
			m_type = NETWORK_STATUS_DATA;

		return false;
	}

	// User data
	else if (m_buffer[6] == 0x73 && m_buffer[7] == 0x12 && m_buffer[10] == 0x20) {
		m_inSeq1 = m_buffer[4] * 256U + m_buffer[5];
		wxUint16 id = m_buffer[14] * 256U + m_buffer[15];

		if (m_buffer[16] == 0x80 && m_inId == 0U) {
			// Only process user headers with the first flag equal to 0x00, 0x01, 0x08, 0x20, 0x28 or 0x40
			if (m_buffer[17] == 0x00 || m_buffer[17] == 0x01 || m_buffer[17] == 0x08 ||
			    m_buffer[17] == 0x20 || m_buffer[17] == 0x28 || m_buffer[17] == 0x40) {
				m_inId = id;					// Take the stream id
				m_type = NETWORK_USER_HEADER;
				return false;
			} else {
				return true;
			}
		}

		// We're not listening for headers, throw it away
		if (m_buffer[16] == 0x80 && m_inId != 0U)
			return true;

		// Check that the stream id matches the valid header, reject otherwise
		if (id == m_inId) {
			// Is this the last packet in the stream?
			if ((m_buffer[16] & 0x40) == 0x40)
				m_inId = 0U;

			m_userSeq2 = m_buffer[16] & 0x0F;
			m_type = NETWORK_USER_DATA;
			return false;
		} else {
			return true;
		}
	}

	// An unknown type
	CUtils::dump(wxT("Unknown packet type from Gateway"), m_buffer, m_length);

	return true;
}

CHeaderData* COpenG2ProtocolHandler::readHeader()
{
	if (m_type != NETWORK_USER_HEADER && m_type != NETWORK_STATUS_HEADER)
		return NULL;

	unsigned int dataLen = m_buffer[8] * 256U + m_buffer[9];

	// Header is the wrong length!
	if (dataLen != 48U) {		// RADIO_HEADER_LENGTH_BYTES + 7U
		CUtils::dump(wxT("Invalid header length from the Gateway"), m_buffer, dataLen + 17U);
		return NULL;
	}

	// Header checksum testing is enabled
	CHeaderData* header = new CHeaderData(m_buffer + 17, RADIO_HEADER_LENGTH_BYTES, true);

	if (!header->isValid()) {
		CUtils::dump(wxT("Header checksum failure from the Gateway"), m_buffer + 17, RADIO_HEADER_LENGTH_BYTES);
		delete header;
		return NULL;
	}

	return header;
}

unsigned int COpenG2ProtocolHandler::readData(unsigned char* buffer, unsigned int length, bool& end)
{
	if (m_type != NETWORK_USER_DATA && m_type != NETWORK_STATUS_DATA)
		return 0U;

	// Get the data length
	unsigned int dataLen = m_buffer[8] * 256U + m_buffer[9] - 7U;

	// Is our buffer too small?
	if (dataLen > length)
		dataLen = length;

	// Is it the end of the data?
	end = (m_buffer[16] & 0x40) == 0x40;

	// Rebuild the sync vector
	wxUint8 seqNo = m_buffer[16] & 0x1F;
	if (!end && seqNo == 0U) {
		m_buffer[26] = 0x55;
		m_buffer[27] = 0x2D;
		m_buffer[28] = 0x16;
	}

	::memcpy(buffer, m_buffer + 17, dataLen);

	return dataLen;
}

void COpenG2ProtocolHandler::reset()
{
	m_inId = 0U;
}

void COpenG2ProtocolHandler::close()
{
	m_socket.close();
}
