/*
 *   Copyright (C) 2009,2010 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ParrotControllerApp.h"
#include "Version.h"
#include "Logger.h"

#include <wx/config.h>
#include <wx/cmdline.h>
#include <wx/filename.h>

IMPLEMENT_APP(CParrotControllerApp)


const wxChar* NAME_PARAM       = wxT("Parrot Name");
const wxChar* NOLOGGING_SWITCH = wxT("nolog");

CParrotControllerApp::CParrotControllerApp() :
wxApp(),
m_name(),
m_nolog(false),
m_frame(NULL),
m_thread(NULL),
m_config(NULL)
{
}

CParrotControllerApp::~CParrotControllerApp()
{
}

bool CParrotControllerApp::OnInit()
{
	SetVendorName(VENDOR_NAME);

	if (!wxApp::OnInit())
		return false;

	if (!m_nolog) {
		wxString logBaseName = LOG_BASE_NAME;
		if (!m_name.IsEmpty()) {
			logBaseName.Append(wxT("_"));
			logBaseName.Append(m_name);
		}

		wxFileName loggerName(wxFileName::GetHomeDir(), logBaseName, wxT("log"));
		wxLog* log = new CLogger(loggerName.GetFullPath());
		wxLog::SetActiveTarget(log);
	} else {
		wxLog::SetActiveTarget(new wxLogStderr);
	}

	m_config = new CParrotControllerConfig(new wxConfig(APPLICATION_NAME), m_name);

	wxString frameName = APPLICATION_NAME + wxT(" - ");
	if (!m_name.IsEmpty()) {
		frameName.Append(m_name);
		frameName.Append(wxT(" - "));
	}
	frameName.Append(VERSION);

	m_frame = new CParrotControllerFrame(frameName);
	m_frame->Show();

	SetTopWindow(m_frame);

	::wxLogInfo(wxT("Starting ") + APPLICATION_NAME + wxT(" - ") + VERSION);

	createThread();

	return wxApp::OnInit();
}

int CParrotControllerApp::OnExit()
{
	::wxLogInfo(APPLICATION_NAME + wxT(" is exiting"));

	m_thread->kill();
	m_thread->Wait();
	delete m_thread;

	delete m_config;

	return 0;
}

void CParrotControllerApp::OnInitCmdLine(wxCmdLineParser& parser)
{
	parser.AddSwitch(NOLOGGING_SWITCH, wxEmptyString, wxEmptyString, wxCMD_LINE_PARAM_OPTIONAL);
	parser.AddParam(NAME_PARAM, wxCMD_LINE_VAL_STRING, wxCMD_LINE_PARAM_OPTIONAL);

	wxApp::OnInitCmdLine(parser);
}

bool CParrotControllerApp::OnCmdLineParsed(wxCmdLineParser& parser)
{
	if (!wxApp::OnCmdLineParsed(parser))
		return false;

	m_nolog = parser.Found(NOLOGGING_SWITCH);

	if (parser.GetParamCount() > 0U)
		m_name = parser.GetParam(0U);

	return true;
}

#if defined(__WXDEBUG__)
void CParrotControllerApp::OnAssertFailure(const wxChar* file, int line, const wxChar* func, const wxChar* cond, const wxChar* msg)
{
	::wxLogFatalError(wxT("Assertion failed on line %d in file %s and function %s: %s %s"), line, file, func, cond, msg);
}
#endif

CParrotControllerStatusData* CParrotControllerApp::getStatus() const
{
	return m_thread->getStatus();
}

void CParrotControllerApp::getMode(PARROT_MODE& mode, unsigned int& beaconTime, wxString& beaconFileName, unsigned int& turnaroundTime, bool& keepFile) const
{
	m_config->getMode(mode, beaconTime, beaconFileName, turnaroundTime, keepFile);
}

void CParrotControllerApp::setMode(PARROT_MODE mode, unsigned int beaconTime, const wxString& beaconFileName, unsigned int turnaroundTime, bool keepFile) const
{
	m_config->setMode(mode, beaconTime, beaconFileName, turnaroundTime, keepFile);
}

void CParrotControllerApp::getNetwork(wxString& repeaterAddress, unsigned int& repeaterPort, wxString& localAddress, unsigned int& localPort) const
{
	m_config->getNetwork(repeaterAddress, repeaterPort, localAddress, localPort);
}

void CParrotControllerApp::setNetwork(const wxString& repeaterAddress, unsigned int repeaterPort, const wxString& localAddress, unsigned int localPort) const
{
	m_config->setNetwork(repeaterAddress, repeaterPort, localAddress, localPort);
}

void CParrotControllerApp::createThread()
{
	m_thread = new CParrotControllerThread;

	PARROT_MODE mode;
	unsigned int beaconTime, turnaroundTime;
	wxString beaconFileName;
	bool keepFile;
	getMode(mode, beaconTime, beaconFileName, turnaroundTime, keepFile);
	m_thread->setMode(mode, beaconTime, beaconFileName, turnaroundTime, keepFile);
	::wxLogInfo(wxT("Mode set to %d, beacon time set to %u secs, beacon file name set to %s, turnaround time set to %u secs, keep file set to %d"), mode, beaconTime, beaconFileName.c_str(), turnaroundTime, keepFile);

	wxString repeaterAddress, localAddress;
	unsigned int repeaterPort, localPort;
	getNetwork(repeaterAddress, repeaterPort, localAddress, localPort);
	::wxLogInfo(wxT("Repeater set to %s:%u, local set to %s:%u"), repeaterAddress.c_str(), repeaterPort, localAddress.c_str(), localPort);

	if (!repeaterAddress.IsEmpty())
		m_thread->setNetwork(repeaterAddress, repeaterPort, localAddress, localPort);

	m_thread->Create();
	m_thread->SetPriority(100U);
	m_thread->Run();
}
